#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import math
_USE_PIL = False
try:
    import gd
except ImportError:
    _USE_PIL = True
    from PIL import Image

BLACK = (0, 0, 0)
BLUE = (0, 0, 197)
RED = (197, 0, 0)
MAGENTA = (197, 0, 197)
GREEN = (0, 198, 0)
CYAN = (0, 198, 197)
YELLOW = (197, 198, 0)
WHITE = (205, 198, 205)
BRIGHT_BLACK = (0, 0, 0)
BRIGHT_BLUE = (0, 0, 255)
BRIGHT_RED = (255, 0, 0)
BRIGHT_MAGENTA = (255, 0, 255)
BRIGHT_GREEN = (0, 255, 0)
BRIGHT_CYAN = (0, 255, 255)
BRIGHT_YELLOW = (255, 255, 0)
BRIGHT_WHITE = (255, 255, 255)
TRANSPARENT = (0, 254, 0)

class Udg:
    def __init__(self, attr, data, mask=None):
        self.attr = attr
        self.data = data
        self.mask = mask
        # If ink or paper is set, it will override attr (see
        # ImageWriter._get_pixel_array())
        self.ink = None
        self.paper = None

_BLANK_UDG = Udg(0, [0] * 8)

def get_snapshot(fname):
    ext = fname[-4:]
    if ext not in ('.sna', '.z80'):
        sys.stderr.write("%s: Unknown file type '%s'\n" % (fname, ext[1:]))
        sys.exit(1)
    f = open(fname, 'rb')
    data = [ord(c) for c in f.read()]
    f.close()
    if ext == '.sna':
        mem = [0] * 16384 + data[27:]
    else:
        version = 1 if sum(data[6:8]) > 0 else 2
        header_size = 30 if version == 1 else 32 + data[30]
        header = data[:header_size]
        if version == 1 and header[12] & 32 == 0:
            ram = data[header_size:-4]
        else:
            ram = _decompress(data[header_size:], version)
        mem = [0] * 16384 + ram
    if len(mem) != 65536:
        sys.stderr.write("ERROR: Memory size is %i\n" % len(mem))
        sys.exit(1)
    return mem

def _decompress_block(ramz):
    block = []
    i = 0
    while i < len(ramz):
        b = ramz[i]
        i += 1
        if b == 237:
            c = ramz[i]
            i += 1
            if c == 237:
                length, byte = ramz[i], ramz[i + 1]
                if length == 0:
                    sys.stderr.write("Found ED ED %x %x\n" % (length, byte))
                    sys.exit(1)
                block += [byte] * length
                i += 2
            else:
                block += [b, c]
        else:
            block.append(b)
    return block

def _decompress(ramz, version):
    if version == 1:
        return _decompress_block(ramz)
    blocks = {}
    j = 0
    while j < len(ramz):
        length = ramz[j] + 256 * ramz[j + 1]
        if length == 65535:
            blocks[ramz[j + 2]] = ramz[j + 3:j + 16387]
            j += 16387
        else:
            blocks[ramz[j + 2]] = _decompress_block(ramz[j + 3:j + 3 + length])
            j += 3 + length
    return blocks[8] + blocks[4] + blocks[5]

class SkoolMemory:
    def __init__(self, snafile, custom):
        self.snapshot = get_snapshot(snafile)
        self.custom = custom
        self._fix_graphic_glitches()
        self._snapshots = [(self.snapshot, '')]

    def build_sprite(self, state, attr=None):
        udg_array = []
        for row in range(4):
            udg_array.append([self.get_sprite_udg(state, attr, row, col) for col in range(3)])
        return udg_array

    def _get_sprite_udg(self, state, attr, ref_page, udg_page):
        ref_addr = state + 256 * ref_page
        ref = self.snapshot[ref_addr]
        if ref == 0:
            udg = [0] * 8
            mask = [255] * 8
        else:
            udg_addr = ref + 256 * udg_page
            udg = self.snapshot[udg_addr:udg_addr + 4096:512]
            mask = self.snapshot[udg_addr + 256:udg_addr + 4352:512]
        return Udg(attr, udg, mask)

    def get_skool_udgs(self, x, y, w, h):
        skool_udgs = []
        for row in range(y, y + h):
            skool_udgs.append([])
            for col in range(x, x + w):
                skool_udgs[-1].append(self.get_skool_udg(row, col))
        return skool_udgs

    def get_logo(self):
        return self.get_box_graphic(self.logo_address)

    def get_score_box_udgs(self):
        return self.get_box_graphic(self.score_box_address)

    def _get_box_udg(self, address, row, col, attr=None):
        if attr is None:
            attr = self.snapshot[address + 8 * row + col]
        start = address + 24 + 64 * row + col
        return Udg(attr, self.snapshot[start:start + 64:8])

    def get_box_graphic(self, address, attr=None):
        udg_array = []
        for row in range(3):
            udg_array.append([self._get_box_udg(address, row, col, attr) for col in range(8)])
        return udg_array

    def get_font_bitmap(self, character):
        base_address = ord(character) + self.font_address + 256
        width = self.snapshot[base_address - 256]
        return self.snapshot[base_address:base_address + 256 * width:256]

    def get_text(self, words):
        columns = []
        for character in words:
            columns.append(0)
            columns.extend(self.get_font_bitmap(character))
        udg_array = []
        for col in range(len(columns)):
            col_byte = columns[col]
            udg_index = col // 8
            bit = int(math.pow(2, 7 - (col - 8 * udg_index)))
            if udg_index == len(udg_array):
                udg_array.append([0] * 8)
            udg = udg_array[udg_index]
            for b in range(8):
                udg[7 - b] |= bit * (col_byte & 1)
                col_byte /= 2
        return udg_array

    def pop_snapshot(self):
        self.snapshot = self._snapshots.pop()[0]

    def push_snapshot(self, name=''):
        self._snapshots.append((self.snapshot[:], name))

class SDMemory(SkoolMemory):
    def __init__(self, snafile, custom):
        SkoolMemory.__init__(self, snafile, custom)
        self.font_address = 55552
        self.logo_address = 61184
        self.score_box_address = 60928

    def get_skool_udg(self, y, x):
        ref_page = y + 152
        attr_addr = x + 128 + 256 * ref_page
        ref_addr = x + 256 * ref_page
        ref = self.snapshot[ref_addr]
        attr = self.snapshot[attr_addr]
        udg_page = 128 + 8 * (x // 32)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        return Udg(attr, udg_bytes)

    def get_sprite_udg(self, state, attr, row, col):
        udg_page = 201 if 71 < (state & 127) < 104 else 185
        ref_page = 173 + col * 4 + row
        return self._get_sprite_udg(state, attr, ref_page, udg_page)

    def get_speech_bubble(self, attr=56):
        bubble_udgs = [[], []]
        for n in range(8):
            bubble_udgs[0].append(Udg(attr, self.snapshot[33008 + n:35056 + n:256]))
            bubble_udgs[1].append(Udg(attr, self.snapshot[33016 + n:35064 + n:256]))
        return bubble_udgs

    def get_bubble_udgs(self):
        attr = 56
        udgs = self.get_speech_bubble(attr)
        udgs[0].append(Udg(attr, self.snapshot[33007:35055:256]))
        udgs[1].append(_BLANK_UDG)
        return udgs

    def get_mutable_udgs(self):
        udgs = []
        for x, y in ((6, 2), (8, 2), (47, 2), (50, 2), (78, 2), (45, 8), (50, 8), (65, 14), (67, 14)):
            shield_udg = self.get_skool_udg(y, x)
            udgs.append(shield_udg)
            udgs.append(self.invert(shield_udg))
        safe_udg = self.get_skool_udg(9, 10)
        udgs.append(safe_udg)
        udgs.append(self.invert(safe_udg))
        return [udgs]

    def invert(self, udg):
        bright = udg.attr & 64
        ink = udg.attr & 7
        paper = (udg.attr & 56) // 8
        attr = bright + ink * 8 + paper
        return Udg(attr, udg.data)

    def _fix_graphic_glitches(self):
        self.snapshot[61009] = 204 # Score box
        self.snapshot[47649] = 252 # EINSTEIN's head (facing left)
        self.snapshot[47777] = 63  # EINSTEIN's head (facing right)
        self.snapshot[47653] = 255 # EINSTEIN's head (lying down, facing left)
        self.snapshot[47781] = 255 # EINSTEIN's head (lying down, facing right)
        self.snapshot[47665] = 255 # BOY WANDER's head (facing left)
        self.snapshot[47793] = 255 # BOY WANDER's head (facing right)
        self.snapshot[53863] = 171 # MR CREAK's ear (facing left)
        self.snapshot[53991] = 213 # MR CREAK's ear (facing right)
        self.snapshot[54367] = 228 # MR WACKER's trousers (facing left)
        self.snapshot[54495] = 39  # MR WACKER's trousers (facing right)
        self.snapshot[53810] = 95  # MR WITHIT's shoes (1) (facing left)
        self.snapshot[54322] = 95  # MR WITHIT's shoes (2) (facing left)
        self.snapshot[54834] = 223 # MR WITHIT's shoes (3) (facing left)
        self.snapshot[54837] = 250 # MR WITHIT's shoes (4) (facing left)
        self.snapshot[53938] = 250 # MR WITHIT's shoes (1) (facing right)
        self.snapshot[54450] = 250 # MR WITHIT's shoes (2) (facing right)
        self.snapshot[54962] = 251 # MR WITHIT's shoes (3) (facing right)
        self.snapshot[54965] = 95  # MR WITHIT's shoes (4) (facing right)
        self.snapshot[51767] = 235 # MR WITHIT's hand (facing left)
        self.snapshot[51895] = 215 # MR WITHIT's hand (facing right)
        self.snapshot[50699] = 222 # Main kids' shoes (facing left)
        self.snapshot[50827] = 123 # Main kids' shoes (facing right)
        self.snapshot[50814] = 222 # Little kids' shoes (facing left)
        self.snapshot[50942] = 123 # Little kids' shoes (facing right)

    def get_play_area_udgs(self):
        return self.get_skool_udgs(0, 0, 96, 21)

    def get_logo_udgs(self):
        logo_udgs = self.get_logo()
        if self.custom == 1:
            self._adjust_logo(logo_udgs)
        return logo_udgs

    def _adjust_logo(self, logo_udgs):
        take = (
            ('01111100', '11001001', '01111000'),
            ('00010001', '00101010', '01000000'),
            ('00010001', '00101100', '01111000'),
            ('00010001', '11101010', '01000000'),
            ('00010001', '00101001', '01111000')
        )
        for i, p_row in enumerate(take):
            for j, p in enumerate(p_row):
                logo_udgs[0][j].data[i + 1] = int(p, 2)
        too = (
            ('11111001', '10001100'),
            ('00100010', '01010010'),
            ('00100010', '01010010'),
            ('00100010', '01010010'),
            ('00100001', '10001100')
        )
        for i, p_row in enumerate(too):
            for j, p in enumerate(p_row):
                logo_udgs[0][j + 5].data[i + 1] = int(p, 2)

class BTSMemory(SkoolMemory):
    def __init__(self, snafile, custom):
        SkoolMemory.__init__(self, snafile, custom)
        self.font_address = 55040
        self.logo_address = 59176
        self.score_box_address = 58920

    def get_skool_udg(self, y, x):
        q_addr = x + 46336
        q = self.snapshot[q_addr]
        h = y + 160
        hl = 180 + q // 2 + 256 * h
        i = 0
        if q & 1:
            p = self.snapshot[hl] & 15
        else:
            p = (self.snapshot[hl] & 240) // 16
        if p == 0:
            i = q & 3 or 2
            if q & 1:
                p = self.snapshot[hl + 256] & 15
            else:
                p = (self.snapshot[hl + 256] & 240) // 16
        attr = i | (8 * p)
        ref_addr = q + 256 * h
        ref = self.snapshot[ref_addr]
        a = 136 // int(math.pow(2, q & 3)) & self.snapshot[144 + q // 4 + 256 * h]
        udg_page = (8 if a & 15 else 0) + (144 if a & 240 else 128)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        return Udg(attr, udg_bytes)

    def get_sprite_udg(self, state, attr, row, col):
        as_norm = state | 128
        udg_page = 183 if as_norm < 208 else 199
        col = 2 - col if state & 128 else col
        ref_page = 215 + col * 4 + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            self._flip_udg(udg)
        return udg

    def get_udg(self, address, attr):
        return Udg(attr, self.snapshot[address:address + 8])

    def get_speech_bubble(self, attr=56):
        top_row = [39392]
        top_row.extend([39648] * 6)
        top_row.append(39904)
        bottom_row = [40160]
        bottom_row.extend([40416] * 6)
        bottom_row.append(40672)
        bubble_udgs = [[self.get_udg(address, attr) for address in top_row]]
        bubble_udgs.append([self.get_udg(address, attr) for address in bottom_row])
        return bubble_udgs

    def get_bubble_udgs(self):
        attr = 56
        udgs = self.get_speech_bubble(attr)
        udgs[0].append(Udg(attr, self.snapshot[39136:39144]))
        udgs[1].append(_BLANK_UDG)
        return udgs

    def _alter_skool_udgs(self, address):
        coords = []
        while True:
            y = self.snapshot[address]
            if y == 255:
                break
            x = self.snapshot[address + 1]
            coords.append((x, y))
            udg_lsb = self.snapshot[address + 2]
            attr = self.snapshot[address + 3]
            d = y + 160
            col_ref = self.snapshot[x + 46336]
            self.snapshot[col_ref + 256 * d] = udg_lsb
            z = int(math.pow(2, col_ref & 3))
            e = 128 + (col_ref + 104) // 2
            a = self.snapshot[e + 256 * d]
            if col_ref & 1:
                a = (a & 240) | (attr & 15)
            else:
                a = (16 * (attr & 15)) | (a & 15)
            self.snapshot[e + 256 * d] = a
            a = attr & 192
            if a & 64:
                a = a - 56
            e = 144 + col_ref // 4
            self.snapshot[e + 256 * d] = self.snapshot[e + 256 * d] & 255 - 136 // z | a // z
            address += 4
        return coords

    def _get_dimensions(self, coords, padding=0):
        x_range = [c[0] for c in coords]
        y_range = [c[1] for c in coords]
        left_x = min(x_range)
        top_y = min(y_range)
        width = max(x_range) - left_x + 1
        height = max(y_range) - top_y + 1
        left_x -= padding
        top_y -= padding
        width += 2 * padding if left_x + width < 192 else padding
        height += 2 * padding if top_y + height < 20 else padding
        return left_x, top_y, width, height

    def get_mutable_udg_array(self, address, padding):
        self.push_snapshot()
        coords = self._alter_skool_udgs(address)
        if padding >= 0:
            udg_array = self.get_skool_udgs(*self._get_dimensions(coords, padding))
        else:
            udg_array = self._get_skool_udgs_by_coords(coords)
        self.pop_snapshot()
        return udg_array

    def get_mutables(self):
        doors = []
        doors.append(56637)    # Left study door
        doors.append(57088)    # Right study door
        doors.append(56576)    # Science Lab storeroom door
        doors.append(57149)    # Boys' skool door
        doors.append(63744)    # Skool gate
        doors.append(56064)    # Drinks cabinet door

        windows = []
        windows.append(55040)  # Top-floor window
        windows.append(55552)  # Middle-floor window

        bike = 57600

        return doors, windows, bike

    def get_cups(self):
        cups = []
        # All cups on the shelf in the boys' skool (empty)
        cups.append((56081, []))
        # Leftmost cup on the shelf in the boys' skool (water/sherry)
        cups.append((56694, [5]))
        # Middle cup on the shelf in the boys' skool (water/sherry)
        cups.append((56950, [5]))
        # Rightmost cup on the shelf in the boys' skool (water/sherry)
        cups.append((57206, [5]))
        # Cup on the shelf in the girls' skool (empty/water/sherry)
        cups.append((56337, [1125, 5]))
        return cups

    def get_inventory_items(self):
        items = []
        items.append(((39144,), 6))        # Safe key
        items.append(((39400,), 7))        # Science Lab storeroom key
        items.append(((39656,), 4))        # Frog
        items.append(((39912, 40168), 5))  # Water pistol (water)
        items.append(((39912, 40168), 3))  # Water pistol (sherry)
        items.append(((40424,), 7))        # Three stinkbombs
        items.append(((40680,), 7))        # Two stinkbombs
        items.append(((40936,), 7))        # One stinkbomb
        items.append(((40928,), 7))        # Mouse
        return items

    def get_mutable_udgs(self):
        udgs = [[None] * 18 for y in range(14)]
        doors, windows, bike = self.get_mutables()

        left_study_door = doors[0]
        left_study_door_closed = self.get_mutable_udg_array(left_study_door, 0)
        self.blit_udgs(udgs, left_study_door_closed, 0, 0)
        left_study_door_open = self.get_mutable_udg_array(left_study_door + 256, 0)
        self.blit_udgs(udgs, left_study_door_open, 3, 0)

        right_study_door = doors[1]
        right_study_door_closed = self.get_mutable_udg_array(right_study_door, 0)
        self.blit_udgs(udgs, right_study_door_closed, 6, 0)
        right_study_door_open = self.get_mutable_udg_array(right_study_door + 256, 0)
        self.blit_udgs(udgs, right_study_door_open, 9, 0)

        science_lab_door = doors[2]
        science_lab_door_closed = self.get_mutable_udg_array(science_lab_door, 0)
        self.blit_udgs(udgs, science_lab_door_closed, 12, 0)
        science_lab_door_open = self.get_mutable_udg_array(science_lab_door + 256, 0)
        self.blit_udgs(udgs, science_lab_door_open, 15, 0)

        skool_door = doors[3]
        skool_door_closed = self.get_mutable_udg_array(skool_door, 0)
        self.blit_udgs(udgs, skool_door_closed, 0, 5)
        skool_door_open = self.get_mutable_udg_array(skool_door + 256, 0)
        self.blit_udgs(udgs, skool_door_open, 3, 5)

        gate = doors[4]
        gate_closed = self.get_mutable_udg_array(gate, 0)
        self.blit_udgs(udgs, gate_closed, 6, 5)
        gate_open = self.get_mutable_udg_array(gate + 256, 0)
        self.blit_udgs(udgs, gate_open, 10, 5)

        cabinet_door = doors[5]
        cabinet_door_closed = self.get_mutable_udg_array(cabinet_door, 0)
        self.blit_udgs(udgs, cabinet_door_closed, 14, 5)
        cabinet_door_open = self.get_mutable_udg_array(cabinet_door + 256, 0)
        self.blit_udgs(udgs, cabinet_door_open, 16, 5)

        top_window = windows[0]
        top_window_closed = self.get_mutable_udg_array(top_window, 0)
        self.blit_udgs(udgs, top_window_closed, 0, 10)
        top_window_open = self.get_mutable_udg_array(top_window + 256, 0)
        self.blit_udgs(udgs, top_window_open, 3, 10)

        middle_window = windows[1]
        middle_window_closed = self.get_mutable_udg_array(middle_window, 0)
        self.blit_udgs(udgs, middle_window_closed, 6, 10)
        middle_window_open = self.get_mutable_udg_array(middle_window + 256, 0)
        self.blit_udgs(udgs, middle_window_open, 8, 10)

        tree_without_bike = self.get_mutable_udg_array(bike, 0)
        self.blit_udgs(udgs, tree_without_bike, 10, 10)
        tree_with_bike = self.get_mutable_udg_array(bike + 256, 0)
        self.blit_udgs(udgs, tree_with_bike, 13, 10)

        cups = self.get_cups()

        boy_cups_empty_udgs = self.get_mutable_udg_array(cups[0][0], 0)
        left_cup_empty_udg = boy_cups_empty_udgs[0][0]
        self.blit_udgs(udgs, [[left_cup_empty_udg]], 14, 7)
        middle_cup_empty_udg = boy_cups_empty_udgs[0][2]
        self.blit_udgs(udgs, [[middle_cup_empty_udg]], 14, 8)
        girl_cup_empty = self.get_mutable_udg_array(cups[4][0], 0)
        self.blit_udgs(udgs, girl_cup_empty, 14, 9)

        left_cup_water = self.get_mutable_udg_array(cups[1][0], 0)
        self.blit_udgs(udgs, left_cup_water, 15, 7)
        middle_cup_water = self.get_mutable_udg_array(cups[2][0], 0)
        self.blit_udgs(udgs, middle_cup_water, 15, 8)
        girl_cup_water = self.get_mutable_udg_array(cups[4][0] + cups[4][1][0], 0)
        self.blit_udgs(udgs, girl_cup_water, 15, 9)

        left_cup_sherry = self.get_mutable_udg_array(cups[1][0] + cups[1][1][0], 0)
        self.blit_udgs(udgs, left_cup_sherry, 16, 7)
        middle_cup_sherry = self.get_mutable_udg_array(cups[2][0] + cups[2][1][0], 0)
        self.blit_udgs(udgs, middle_cup_sherry, 16, 8)
        girl_cup_sherry = self.get_mutable_udg_array(cups[4][0] + sum(cups[4][1]), 0)
        self.blit_udgs(udgs, girl_cup_sherry, 16, 9)

        for y in range(len(udgs)):
            for x in range(len(udgs[y])):
                if udgs[y][x] is None:
                    udgs[y][x] = _BLANK_UDG

        if self.custom == 1:
            self._draw_shields_and_safe(udgs)
        return udgs

    def get_inventory_udgs(self):
        udgs = []
        for addresses, attr in self.get_inventory_items():
            for address in addresses:
                udgs.append(self.get_udg(address, attr))
        return [udgs]

    def blit_udgs(self, montage, udgs, x, y):
        for i, row in enumerate(udgs):
            montage[y + i][x:x + len(row)] = row

    def _fix_graphic_glitches(self):
        self.snapshot[59001] = 204 # Score box
        self.snapshot[47249] = 252 # EINSTEIN's head
        self.snapshot[47254] = 255 # EINSTEIN's head (lying down)
        self.snapshot[47265] = 255 # BOY WANDER's head
        self.snapshot[53356] = 171 # MR CREAK's ear
        self.snapshot[53896] = 228 # MR WACKER's trousers
        self.snapshot[53461] = 95  # MR WITHIT's and ALBERT's shoes (1)
        self.snapshot[53973] = 95  # MR WITHIT's and ALBERT's shoes (2)
        self.snapshot[54485] = 223 # MR WITHIT's and ALBERT's shoes (3)
        self.snapshot[54487] = 250 # MR WITHIT's and ALBERT's shoes (4)
        self.snapshot[51417] = 235 # MR WITHIT's and MISS TAKE's hand
        self.snapshot[50281] = 222 # Kids' shoes

    def get_shield_udgs(self):
        return (self.get_skool_udg(2, 74), self.get_skool_udg(2, 79), self.get_skool_udg(2, 82))

    def get_safe_udg(self):
        return self.get_skool_udg(2, 81)

    def get_play_area_udgs(self):
        skool_udgs = self.get_skool_udgs(0, 0, 192, 21)
        if self.custom == 1:
            self._place_shields(skool_udgs)
        return skool_udgs

    def _place_shields(self, skool_udgs):
        self.s_udgs = []

        # Add shield UDGs with green, cyan, yellow and white paper (ink
        # unchanged)
        shield_udgs = self.get_shield_udgs()
        for paper in (4, 5, 6, 7):
            udgs = []
            for shield_udg in shield_udgs:
                ink = shield_udg.attr & 7
                udg = Udg(8 * paper + ink, shield_udg.data)
                udgs.append(udg)
            self.s_udgs.append(udgs)
        self.s_udgs.append([self.get_safe_udg()])

        # Remove the picture from the wall in the Science Lab storeroom
        for x in (55, 56, 57):
            for y in (9, 10):
                skool_udgs[y][x].data = [0] * 8
        skool_udgs[9][58].data = [3] * 8
        skool_udgs[10][58].data = (3, 3, 3, 3, 3, 0, 0, 0)

        # Place the extra shields (x, y, paper [4567], index [012])
        shields = (
            (4, 2, 5, 0), (9, 2, 5, 1), (162, 2, 5, 1), (167, 2, 5, 2),
            (33, 8, 6, 1), (38, 8, 6, 2), (57, 9, 5, 2), (173, 8, 6, 0),
            (4, 15, 4, 1), (8, 15, 4, 2), (12, 15, 4, 0), (30, 14, 7, 2),
            (186, 14, 7, 1)
        )
        for x, y, udg_i, udg_j in shields:
            s_udg = self.s_udgs[udg_i - 4][udg_j]
            skool_udgs[y][x] = Udg(s_udg.attr, s_udg.data[:])

    def _draw_shields_and_safe(self, mutable_udgs):
        # Remove the picture on the wall from the storeroom door images
        for x in (13, 14):
            for y in (0, 1):
                mutable_udgs[y][x].data = [0] * 8
        mutable_udgs[0][17].data[6:] = [3, 3]
        mutable_udgs[1][17].data[:4] = [3, 3, 3, 3]

        # Draw the shields and the safe
        udg_locations = (
            ( 6,  9), ( 8,  9), (10,  9),  # Shields with green paper
            (16, 10), (16, 11), (16, 12),  # Shields with cyan paper
            ( 0, 13), ( 2, 13), ( 4, 13),  # Shields with yellow paper
            (10, 13), (12, 13), (14, 13),  # Shields with white paper
            (12,  9)                       # Safe
        )
        i = 0
        for udg_set in self.s_udgs:
            for udg in udg_set:
                ink = udg.attr & 7
                paper = (udg.attr & 56) // 8
                bright = udg.attr & 64
                inverse_udg = Udg(bright + 8 * ink + paper, udg.data)
                x, y = udg_locations[i]
                mutable_udgs[y][x] = udg
                mutable_udgs[y][x + 1] = inverse_udg
                i += 1

    def get_logo_udgs(self):
        logo_udgs = self.get_logo()
        if self.custom == 1:
            self._adjust_logo(logo_udgs)
        return logo_udgs

    def _adjust_logo(self, logo_udgs):
        # Shift logo 8 pixels to the left to make room for 'daze'
        for row in logo_udgs:
            row.append(row.pop(0))

        # Write the word 'daze' on the logo
        bit_patterns = (
            ('',         '00000000', '00000000'),
            ('',         '00000000', '00000000'),
            ('',         '00000000', '00000000'),
            ('',         '00000000', '00011100'),
            ('',         '00000000', '00100010'),
            ('',         '00000000', '00100100'),
            ('',         '00000000', '00101001'),
            ('',         '00000000', '10010010'),
            ('10000100', '00000001', '10001100'),
            ('10001000', '00000010', '10000000'),
            ('01110000', '00000100', '10010000'),
            ('00000000', '00000000', '10100000'),
            ('10000000', '01110000', '11000000'),
            ('10000000', '10001000', '10000000'),
            ('00100000', '10000100', '00000000'),
            ('00010000', '10000110', '00000000'),
            ('00001000', '10000100', '00000000'),
            ('00000100', '01001000', '00000000'),
            ('00011110', '00110000', '00000000'),
            ('00010001', '00000000', '00000000'),
            ('00010000', '10000000', '00000000'),
            ('00010001', '00000000', '00000000'),
            ('00010001', '00000000', '00000000'),
            ('00001110', '00000000', '00000000')
        )

        # Convert the bit patterns into UDG data
        new_udgs = []
        for i, patterns in enumerate(bit_patterns):
            row = i // 8
            if len(new_udgs) == row:
                new_udgs.append([])
                for j in range(len(patterns)):
                    new_udgs[row].append([])
            for k, p in enumerate(patterns):
                if p:
                    new_udgs[row][k].append(int(p, 2))

        # Insert the UDGs
        for y in range(3):
            for x in range(1, len(new_udgs[y]) + 1):
                data = new_udgs[y][-x]
                if data:
                    logo_udgs[y][-x].data = data

class ImageWriter:
    def write_image(self, udgs, fname, scale, mask, width, flip):
        if flip:
            udgs = self._flip_udgs(udgs)
        self._write_image(udgs, fname, scale, mask, width)

    def _open_file(self, fname):
        return open(fname, 'wb')

    def _get_pixel_array(self, udg_array, mask):
        # Convert UDG array into byte array
        byte_array = []
        for row in udg_array:
            for line in range(8):
                byte_array.append([(udg.ink, udg.paper, udg.attr, udg.data[line], udg.mask[line] if mask else 0) for udg in row])

        # Convert byte array into pixel array
        pixel_array = []
        for row in byte_array:
            pixel_array.append([])
            for ink, paper, attr, byte, mask_byte in row:
                bright = (attr & 64) // 8
                ink = ink or (bright + (attr & 7))
                paper = paper or (bright + (attr & 56) // 8)
                for b in range(8):
                    colour = paper
                    pixel = byte & 128
                    if mask:
                        if pixel & mask_byte:
                            colour = ink
                        elif mask_byte & 128:
                            colour = 16
                        mask_byte *= 2
                    elif pixel:
                        colour = ink
                    pixel_array[-1].append(colour)
                    byte *= 2

        return pixel_array

    def _gd_create_colours(self, image):
        colours = []
        colours.append(image.colorAllocate(BLACK))
        colours.append(image.colorAllocate(BLUE))
        colours.append(image.colorAllocate(RED))
        colours.append(image.colorAllocate(MAGENTA))
        colours.append(image.colorAllocate(GREEN))
        colours.append(image.colorAllocate(CYAN))
        colours.append(image.colorAllocate(YELLOW))
        colours.append(image.colorAllocate(WHITE))
        colours.append(colours[0])
        colours.append(image.colorAllocate(BRIGHT_BLUE))
        colours.append(image.colorAllocate(BRIGHT_RED))
        colours.append(image.colorAllocate(BRIGHT_MAGENTA))
        colours.append(image.colorAllocate(BRIGHT_GREEN))
        colours.append(image.colorAllocate(BRIGHT_CYAN))
        colours.append(image.colorAllocate(BRIGHT_YELLOW))
        colours.append(image.colorAllocate(BRIGHT_WHITE))
        colours.append(image.colorAllocate(TRANSPARENT))
        return colours

    def _gd_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        pixel_array = self._get_pixel_array(udgs, mask)
        width = width or len(pixel_array[0])
        image = gd.image((scale * width, scale * len(pixel_array)))
        colours = self._gd_create_colours(image)
        for y, row in enumerate(pixel_array):
            for x, attr in enumerate(row[:width]):
                x1, y1 = x * scale, y * scale
                x2, y2 = x1 + scale - 1, y1 + scale - 1
                image.filledRectangle((x1, y1), (x2, y2), colours[attr])
        f = self._open_file(img_file)
        image.writePng(f)
        f.close()

    def _pil_create_colours(self):
        self.colours = []
        self.colours.extend(BLACK)
        self.colours.extend(BLUE)
        self.colours.extend(RED)
        self.colours.extend(MAGENTA)
        self.colours.extend(GREEN)
        self.colours.extend(CYAN)
        self.colours.extend(YELLOW)
        self.colours.extend(WHITE)
        self.colours.extend(BRIGHT_BLACK)
        self.colours.extend(BRIGHT_BLUE)
        self.colours.extend(BRIGHT_RED)
        self.colours.extend(BRIGHT_MAGENTA)
        self.colours.extend(BRIGHT_GREEN)
        self.colours.extend(BRIGHT_CYAN)
        self.colours.extend(BRIGHT_YELLOW)
        self.colours.extend(BRIGHT_WHITE)
        self.colours.extend(TRANSPARENT)

    def _pil_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        pixel_array = self._get_pixel_array(udgs, mask)
        width = width or len(pixel_array[0])
        image = Image.new('P', (scale * width, scale * len(pixel_array)), None)
        image.putpalette(self.colours)
        pixels = []
        for row in pixel_array:
            pixel_row = []
            for pixel in row[:width]:
                pixel_row.extend([pixel] * scale)
            for i in range(scale):
                pixels.extend(pixel_row)
        image.putdata(pixels)
        f = self._open_file(img_file)
        image.save(f, 'PNG', optimize=True)
        f.close()

    _write_image = _pil_write_image if _USE_PIL else _gd_write_image

    def _flip_udgs(self, udgs):
        flipped = []
        for row in udgs:
            flipped_row = []
            for udg in row:
                flipped_data = []
                flipped_mask = []
                for i in range(8):
                    flipped_data.append(self._flip_byte(udg.data[i]))
                    if udg.mask:
                        flipped_mask.append(self._flip_byte(udg.mask[i]))
                flipped_udg = Udg(udg.attr, flipped_data, flipped_mask)
                flipped_udg.ink = udg.ink
                flipped_udg.paper = udg.paper
                flipped_row.insert(0, flipped_udg)
            flipped.append(flipped_row)
        return flipped

    def _flip_byte(self, byte):
        f = 0
        for b in range(8):
            f = 2 * f + (byte & 1)
            byte /= 2
        return f
